
import pandas as pd
import numpy as np
import pickle5 as pickle
import json

from lapsolver import solve_dense
from mot_neural_solver_bub.utils.iou import iou
from mot_neural_solver_bub.path_cfg import  TO_ADD_PATH
import os

##########################################################################################
# Classes used to store and process detections for every sequence
##########################################################################################

class DataFrameWSeqInfo(pd.DataFrame):
    """
    Class used to store each sequences's processed detections as a DataFrame. We just add a metadata atribute to
    pandas DataFrames it so that sequence metainfo such as fps, etc. can be stored in the attribute 'seq_info_dict'.
    This attribute survives serialization.
    This solution was adopted from:
    https://pandas.pydata.org/pandas-docs/stable/development/extending.html#define-original-properties
    """
    _metadata = ['seq_info_dict']

    @property
    def _constructor(self):
        return DataFrameWSeqInfo

class MOTSeqProcessor:
    """
    Class to process detections files coming from different mot_seqs.
    Main method is process_detections. It does the following:
    - Loads a DataFrameWSeqInfo (~pd.DataFrame) from a  detections file (self.det_df) via a the 'det_df_loader' func
    corresponding to the sequence type (mapped via _SEQ_TYPES)
    - Adds Sequence Info to the df (fps, img size, moving/static camera, etc.) as an additional attribute (_get_det_df)
    - If GT is available, assigns GT identities to the detected boxes via bipartite matching (_assign_gt)
    - Stores the df on disk (_store_det_df)
    - If required, precomputes CNN embeddings for every detected box and stores them on disk (_store_embeddings)

    """
    def __init__(self, dataset_path, seq_name, dataset_params, logger = None):
        self.seq_name = seq_name
        self.dataset_path = dataset_path
        self.dataset_params = dataset_params
        self.logger = logger

    def _ensure_boxes_in_frame(self):
        """
        Determines whether boxes are allowed to have some area outside the image (all GT annotations in MOT15 are inside
        the frame hence we crop its detections to also be inside it)
        """
        initial_bb_top = self.det_df['bb_top'].values.copy()
        initial_bb_left = self.det_df['bb_left'].values.copy()

        self.det_df['bb_top'] = np.maximum(self.det_df['bb_top'].values, 0).astype(int)
        self.det_df['bb_left'] = np.maximum(self.det_df['bb_left'].values, 0).astype(int)

        bb_top_diff = self.det_df['bb_top'].values - initial_bb_top
        bb_left_diff = self.det_df['bb_left'].values - initial_bb_left

        self.det_df['bb_height'] -= bb_top_diff
        self.det_df['bb_width'] -= bb_left_diff

        img_height, img_width = self.det_df.seq_info_dict['frame_height'], self.det_df.seq_info_dict['frame_width']
        self.det_df['bb_height'] = np.minimum(img_height - self.det_df['bb_top'], self.det_df['bb_height']).astype(int)
        self.det_df['bb_width'] = np.minimum(img_width - self.det_df['bb_left'], self.det_df['bb_width']).astype(int)

    def _assign_gt(self):
        """
        Assigns a GT identity to every detection in self.det_df, based on the ground truth boxes in self.gt_df.
        The assignment is done frame by frame via bipartite matching.
        """
        if self.det_df.seq_info_dict['has_gt'] and not self.det_df.seq_info_dict['is_gt']:
            print(f"Assigning ground truth identities to detections to sequence {self.seq_name}")
            for frame in self.det_df['frame'].unique():
                frame_detects = self.det_df[self.det_df.frame == frame]
                frame_gt = self.gt_df[self.gt_df.frame == frame]

                # Compute IoU for each pair of detected / GT bounding box
                iou_matrix = iou(frame_detects[['bb_top', 'bb_left', 'bb_bot', 'bb_right']].values,
                                 frame_gt[['bb_top', 'bb_left', 'bb_bot', 'bb_right']].values)

                iou_matrix[iou_matrix < self.dataset_params['gt_assign_min_iou']] = np.nan
                dist_matrix = 1 - iou_matrix
                assigned_detect_ixs, assigned_detect_ixs_ped_ids = solve_dense(dist_matrix)
                unassigned_detect_ixs = np.array(list(set(range(frame_detects.shape[0])) - set(assigned_detect_ixs)))

                assigned_detect_ixs_index = frame_detects.iloc[assigned_detect_ixs].index
                assigned_detect_ixs_ped_ids = frame_gt.iloc[assigned_detect_ixs_ped_ids]['id'].values
                unassigned_detect_ixs_index = frame_detects.iloc[unassigned_detect_ixs].index

                self.det_df.loc[assigned_detect_ixs_index, 'id'] = assigned_detect_ixs_ped_ids
                self.det_df.loc[unassigned_detect_ixs_index, 'id'] = -1  # False Positives

    def _get_det_df(self):
        ph=0
        if str(TO_ADD_PATH)==self.dataset_path:
            self.dataset_path=''
        with open(TO_ADD_PATH+self.dataset_path+self.seq_name+'/Meta.json','r') as f:
            seq_info_dict=json.load(f)

        with open(TO_ADD_PATH+seq_info_dict['seq_path']+"/Nodes.pkl", "rb") as fh:
            self.det_df=pickle.load(fh)
        if os.path.exists(TO_ADD_PATH+seq_info_dict['seq_path']+"/NodesGT.pkl"):
            with open(TO_ADD_PATH+seq_info_dict['seq_path']+"/NodesGT.pkl", "rb") as fh:
                self.gt_df=pickle.load(fh) 
        else:
            self.gt_df=None
        self.det_df = DataFrameWSeqInfo(self.det_df)
        
        self.det_df.seq_info_dict = seq_info_dict

        y=np.stack(self.det_df['feat'].values.tolist())[...,0] 
        x=np.stack(self.det_df['feat'].values.tolist())[...,1] 
        feat=np.square(y-np.expand_dims(self.det_df['posy'].values,axis=1))+np.square(x-np.expand_dims(self.det_df['posx'].values,axis=1))
        self.det_df['feat_out']=list(np.stack(self.det_df['feat'].values.tolist())[...,:2])
        self.det_df['feat']=list(feat/np.max(feat,axis=0))
        
        # Add some additional box measurements that might be used for graph construction
        self.det_df['bb_height'] = (self.det_df['bb_bot'] - self.det_df['bb_top']).values
        self.det_df['bb_width'] = (self.det_df['bb_right'] - self.det_df['bb_left']).values
        self.det_df['det_id']=self.det_df['detection_id'] 
        self.det_df['id'] =self.det_df.pop('detection_id')
        self.det_df['detection_id'] = np.arange(self.det_df.shape[0]) # This id is used for future tastks
        if 'level_0' in self.det_df.columns:
            self.det_df=self.det_df.drop(['level_0'],axis=1)
        
        self.det_df = self.det_df.reset_index().sort_values(by=['frame', 'detection_id']) 
        return self.det_df


    def process_detections(self):
        # See class header
        self._get_det_df()
        if self.gt_df is not None:
            self._assign_gt()

        return self.det_df

    def load_or_process_detections(self):
        seq_det_df=self.process_detections()
        return seq_det_df
