import numpy as np
import pandas as pd

import torch

from mot_neural_solver_bub.data.mot_graph import Graph

from mot_neural_solver_bub.tracker.projectors import GreedyProjector, ExactProjector
from mot_neural_solver_bub.tracker.postprocessing import Postprocessor

from mot_neural_solver_bub.utils.graph import to_undirected_graph, to_lightweight_graph

from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import connected_components
from lapsolver import solve_dense

#VIDEO_COLUMNS = ['frame', 'ped_id','posx','posy','size']
#TRACKING_OUT_COLS = ['frame', 'ped_id','midx','midy','size']
# What cols to overtake from readin
VIDEO_COLUMNS = ['frame', 'ped_id', 'bb_left', 'bb_top', 'bb_width', 'bb_height', 'bb_right', 'bb_bot','posx','posy','size','tN','det_id','Grad','feat_out']
TRACKING_OUT_COLS = ['frame', 'ped_id', 'bb_left', 'bb_top', 'bb_width', 'bb_height','posx','posy','size','tN','det_id','Grad','feat_out','Interpolated']
TRACKING_EVAL_COLS = ['frame', 'ped_id', 'bb_left', 'bb_top', 'bb_width', 'bb_height','posx','posy','size']

VIDEO_COLUMNS_TRAIN = ['frame', 'ped_id', 'bb_left', 'bb_top', 'bb_width', 'bb_height', 'bb_right', 'bb_bot','posx','posy','size','Dia']
TRACKING_OUT_COLS_TRAIN = ['frame', 'ped_id', 'bb_left', 'bb_top', 'bb_width', 'bb_height','posx','posy','size','Dia']
class MPNTracker:
    """
    Class used to track video sequences.

    See 'track'  method for an overview.
    """

    def __init__(self, dataset, graph_model, use_gt, eval_params=None,
                 dataset_params=None, logger=None):

        self.dataset = dataset
        self.use_gt = use_gt
        self.logger = logger

        self.eval_params = eval_params
        self.dataset_params = dataset_params

        self.graph_model = graph_model
        if torch.cuda.is_available():
            self.device=True
        else:
            self.device=False
        if self.graph_model is not None:
            self.graph_model.eval()

    def _estimate_frames_per_graph(self, seq_name):
        """
        Determines the number of frames to be included in each batch of frames evaluated within a sequence
        """
        num_frames = len(self.dataset.seq_det_dfs[seq_name].frame.unique())
        num_detects = self.dataset.seq_det_dfs[seq_name].shape[0]

        avg_detects_per_frame = num_detects / float(num_frames)
        expected_frames_per_graph = round(self.dataset.dataset_params['max_detects'] / avg_detects_per_frame)

        return min(expected_frames_per_graph, self.dataset.dataset_params['frames_per_graph'])

    def _estimate_max_frame_dist(self, seq_name, frames_per_graph):
        step_size = self.dataset.seq_info_dicts[seq_name]['step_size']
        # frames_per_graph = self._estimate_frames_per_graph(seq_name)

        # TODO: Should use seconds as unit, and not number of frames
        if self.dataset.dataset_params['max_frame_dist'] == 'max':
            max_frame_dist = step_size * (frames_per_graph - 1)

        else:
            max_frame_dist = self.dataset.dataset_params['max_frame_dist']

        return max_frame_dist

    def _load_full_seq_graph_object(self, seq_name, start_frame, end_frame, max_frame_dist):
        """
        Loads a MOTGraph (see data/mot_graph.py) object corresponding to the entire sequence.
        """

        full_graph = self.dataset.get_from_frame_and_seq(seq_name=seq_name,
                                                         start_frame=start_frame,
                                                         end_frame=end_frame,
                                                         return_full_object=True,
                                                         ensure_end_is_in=True,
                                                         max_frame_dist=max_frame_dist,
                                                         inference_mode=True)
        # full_graph.frames_per_graph = frames_per_graph
        return full_graph

    def _predict_edges(self, subgraph):
        """
        Predicts edge values for a subgraph (i.e. batch of frames) from the entire sequence.
        Args:
            subgraph: Graph Object corresponding to a subset of frames

        Returns:
            tuple containing a torch.Tensor with the predicted value for every edge in the subgraph, and a binary mask
            indicating which edges inside the subgraph where pruned with KNN
        """

        knn_mask=torch.ones(subgraph.edge_labels.shape[0],dtype=torch.bool)
        # Predict active edges
        if self.use_gt:  # For debugging purposes and obtaining oracle results
            pruned_edge_preds = subgraph.edge_labels

        else:
            with torch.no_grad():
                pruned_edge_preds = torch.sigmoid(self.graph_model(subgraph)['classified_edges'][-1].view(-1))

        edge_preds = torch.zeros(knn_mask.shape[0]).to(pruned_edge_preds.device)
        edge_preds[knn_mask] = pruned_edge_preds
        if self.eval_params['set_pruned_edges_to_inactive']:
            return edge_preds, torch.ones_like(knn_mask)

        else:
            return edge_preds, knn_mask  # In this case, pruning an edge counts as not predicting a value for it at all
            # However, if it is pruned for every batch, then it counts as inactive.

    def _evaluate_graph_in_batches(self, subseq_graph, frames_per_graph):
        """
        Feeds the entire sequence though the MPN in batches. It does so by applying a 'sliding window' over the sequence,
        where windows correspond consecutive pairs of start/end frame locations (e.g. frame 1 to 15, 5 to 20, 10 to 25,
        etc.).
        For every window, a subgraph is created by selecting all detections that fall inside it. Then this graph
        is fed to the message passing network, and predictions are stored.
        Since windows overlap, we end up with several predictions per edge. We simply average them overall all
        windows.
        """
        if self.device:
            device = torch.device('gpu')
        else:
            device = torch.device('cpu')
        all_frames = np.array(subseq_graph.frames)
        frame_num_per_node = torch.from_numpy(subseq_graph.graph_df.frame.values).to(device)
        node_names = torch.arange(subseq_graph.graph_obj.x.shape[0])

        # Iterate over overlapping windows of (starg_frame, end_frame)
        overall_edge_preds = torch.zeros(subseq_graph.graph_obj.num_edges).to(device)
        overall_num_preds = overall_edge_preds.clone()
        for eval_round, (start_frame, end_frame) in enumerate(zip(all_frames, all_frames[frames_per_graph - 1:])):
            assert ((start_frame <= all_frames) & (all_frames <= end_frame)).sum() == frames_per_graph

            # Create and evaluate a a subgraph corresponding to a batch of frames
            nodes_mask = (start_frame <= frame_num_per_node) & (frame_num_per_node <= end_frame)
            edges_mask = nodes_mask[subseq_graph.graph_obj.edge_index[0]] & nodes_mask[
                subseq_graph.graph_obj.edge_index[1]]

            subgraph = Graph(x=subseq_graph.graph_obj.x[nodes_mask],
                             edge_attr=subseq_graph.graph_obj.edge_attr[edges_mask],                            
                             edge_index=subseq_graph.graph_obj.edge_index.T[edges_mask].T - node_names[nodes_mask][0])

            if hasattr(subseq_graph.graph_obj, 'edge_labels'):
                subgraph.edge_labels = subseq_graph.graph_obj.edge_labels[edges_mask]

            # Predict edge values for the current batch
            edge_preds, pred_mask = self._predict_edges(subgraph=subgraph)

            # Store predictions
            overall_edge_preds[edges_mask] += edge_preds
            assert (overall_num_preds[torch.where(edges_mask)[0][pred_mask]] == overall_num_preds[edges_mask][
                pred_mask]).all()
            overall_num_preds[torch.where(edges_mask)[0][pred_mask]] += 1

        # Average edge predictions over all batches, and over each pair of directed edges
        final_edge_preds = overall_edge_preds / overall_num_preds
        final_edge_preds[torch.isnan(final_edge_preds)] = 0
        subseq_graph.graph_obj.edge_preds = final_edge_preds
        to_undirected_graph(subseq_graph, attrs_to_update=('edge_preds', 'edge_labels'))
        to_lightweight_graph(subseq_graph)
        return subseq_graph
        # print(time() - t)

    def _project_graph_model_output(self, subseq_graph):
        """
        Rounds MPN predictions either via Linear Programming or a greedy heuristic
        """
        if self.eval_params['rounding_method'] == 'greedy':
            projector = GreedyProjector(subseq_graph)

        elif self.eval_params['rounding_method'] == 'exact':
            projector = ExactProjector(subseq_graph, solver_backend=self.eval_params['solver_backend'])

        else:
            raise RuntimeError("Rounding type for projector not understood")

        projector.project()

        subseq_graph.graph_obj = subseq_graph.graph_obj.numpy()
        subseq_graph.constr_satisf_rate = projector.constr_satisf_rate

        return subseq_graph

    def _assign_ped_ids(self, subseq_graph):
        """
        Assigns pedestrian Ids to each detection in the sequence, by determining all connected components in the graph
        """
        # Only keep the non-zero edges and Express the result as a CSR matrix so that it can be fed to 'connected_components')
        nonzero_mask = subseq_graph.graph_obj.edge_preds == 1
        nonzero_edge_index = subseq_graph.graph_obj.edge_index.T[nonzero_mask].T
        nonzero_edges = subseq_graph.graph_obj.edge_preds[nonzero_mask].astype(int)
        graph_shape = (subseq_graph.graph_obj.num_nodes, subseq_graph.graph_obj.num_nodes)
        csr_graph = csr_matrix((nonzero_edges, (tuple(nonzero_edge_index))), shape=graph_shape)

        # Get the connected Components:
        n_components, labels = connected_components(csgraph=csr_graph, directed=False, return_labels=True)
        assert len(labels) == subseq_graph.graph_df.shape[0], "Ped Ids Label format is wrong"

        # Each Connected Component is a Ped Id. Assign those values to our DataFrame:
        final_projected_output = subseq_graph.graph_df.copy()
        final_projected_output['ped_id'] = labels
        final_projected_output = final_projected_output[VIDEO_COLUMNS + ['detection_id']].copy()

        return final_projected_output

    def _determine_seq_cutpoints(self, seq_name):
        """
        In order to reduce the memory requirements of processing large sequences (i.e. avoid storing the entire graph
        corresponding to the sequence), we split them into smaller sequences, get the tracking output of each of them,
        and then merge the results.

        """
        seq_det_df = self.dataset.seq_det_dfs[seq_name]
        step_size = self.dataset.seq_info_dicts[seq_name]['step_size']
        start_frame, end_frame = seq_det_df.frame.min(), seq_det_df.frame.max()

        valid_frames = np.arange(start_frame, end_frame + 1, step_size)
        if end_frame not in valid_frames:
            valid_frames = valid_frames.tolist() + [end_frame]
        seq_det_df = seq_det_df[seq_det_df.frame.isin(valid_frames)].copy()

        assert_str = f"{self.eval_params['max_dets_per_graph_seq']} to be larger than than twice the maximum amount of dets possible to have in one frame"
        assert self.eval_params['max_dets_per_graph_seq'] > 2 * seq_det_df.groupby('frame')[
            'frame'].count().max(), assert_str

        num_dets = seq_det_df.shape[0]
        frame_cutpoints = [seq_det_df.frame.iloc[i] for i in
                           np.arange(0, num_dets, self.eval_params['max_dets_per_graph_seq'])]
        frame_cutpoints += [seq_det_df.frame.iloc[-1] + 1]
        assert len(frame_cutpoints) == len(set(frame_cutpoints))

        return frame_cutpoints

    def _merge_subseq_dfs(self, subseq_dfs):
        seq_df = subseq_dfs[0]
        for subseq_df in subseq_dfs[1:]:
            # Make sure that ped_ids in subseq_df are new and all greater than the ones in seq_df:
            subseq_df['ped_id'] += seq_df['ped_id'].max() + 1


            intersect_frames = np.intersect1d(seq_df.frame, subseq_df.frame)
            left_df = seq_df[['detection_id', 'ped_id']][seq_df.frame.isin(intersect_frames)]
            left_ids_pos = left_df[['ped_id']].drop_duplicates(); left_ids_pos['ped_id_pos'] = np.arange(left_ids_pos.shape[0])
            left_df = left_df.merge(left_ids_pos, on='ped_id').set_index('detection_id')


            right_df = subseq_df[['detection_id', 'ped_id']][subseq_df.frame.isin(intersect_frames)]
            right_ids_pos = right_df[['ped_id']].drop_duplicates(); right_ids_pos['ped_id_pos'] = np.arange(right_ids_pos.shape[0])
            right_df = right_df.merge(right_ids_pos, on='ped_id').set_index('detection_id')



            # Build the linear assignment matrix
            common_boxes = left_df[['ped_id_pos']].join(right_df['ped_id_pos'], lsuffix='_left', rsuffix='_right').dropna(thresh = 2).reset_index().groupby(['ped_id_pos_left', 'ped_id_pos_right'])['detection_id'].count()
            common_boxes = common_boxes.reset_index().astype(int)

            cost_mat = np.full((common_boxes['ped_id_pos_left'].max() +1 , common_boxes['ped_id_pos_right'].max() + 1), fill_value = np.nan)
            cost_mat[common_boxes['ped_id_pos_left'].values, common_boxes['ped_id_pos_right'].values] =  - common_boxes['detection_id'].values
            matched_left_ids_pos, matched_right_ids_pos = solve_dense(cost_mat)
            matched_ids = pd.DataFrame(data = np.stack((left_ids_pos['ped_id'].values[matched_left_ids_pos],
                                                       right_ids_pos['ped_id'].values[matched_right_ids_pos])).T,
                                       columns = ['left_ped_id', 'right_ped_id'])


            # Assign the ids matched to subseq_df
            subseq_df = pd.merge(subseq_df, matched_ids, how='outer', left_on='ped_id', right_on='right_ped_id')
            subseq_df['left_ped_id'].fillna(np.inf, inplace=True)
            subseq_df['ped_id'] = np.minimum(subseq_df['left_ped_id'], subseq_df['ped_id'])

            # Update seq_df
            seq_df = pd.concat([seq_df, subseq_df[subseq_df['frame'] > seq_df['frame'].max()]])

        return seq_df

    def track(self, seq_name, output_path=None):
        """
        Main method. Given a sequence name, it tracks all detections and produces an output DataFrame, where each
        detection is assigned an ID.

        It starts loading a the graph corresponding to an entire video sequence and detections, then uses an MPN to
        sequentially evaluate batches of frames (i.e. subgraphs) and finally rounds predictions and applies
        postprocessing.

        """
        from time import time

        #print(f"Processing Seq {seq_name}")
        frames_per_graph = self._estimate_frames_per_graph(seq_name)
        max_frame_dist = self._estimate_max_frame_dist(seq_name, frames_per_graph)
        frame_cutpoints = self._determine_seq_cutpoints(seq_name)
        subseq_dfs = []
        constr_sr = 0
        for start_frame, end_frame in zip(frame_cutpoints[:-1], frame_cutpoints[1:]):
            t = time()

            # Load the graph corresponding to the entire subsequence
            subseq_graph = self._load_full_seq_graph_object(seq_name, start_frame if start_frame == frame_cutpoints[0] else start_frame - max_frame_dist,
                                                            end_frame, max_frame_dist)

            # Feed graph through MPN in batches
            subseq_graph = self._evaluate_graph_in_batches(subseq_graph, frames_per_graph)

            # Round predictions and assign IDs to trajectories
            subseq_graph = self._project_graph_model_output(subseq_graph)
            constr_sr += subseq_graph.constr_satisf_rate
            subseq_df = self._assign_ped_ids(subseq_graph)
            subseq_dfs.append(subseq_df)

            del subseq_graph



        constr_sr /= (len(frame_cutpoints) - 1)
        seq_df = self._merge_subseq_dfs(subseq_dfs)
        # Postprocess trajectories
        if self.eval_params['add_tracktor_detects']:

            seq_df = self._add_tracktor_detects(seq_df, seq_name)


        postprocess = Postprocessor(seq_df.copy(),
                                    seq_info_dict=self.dataset.seq_info_dicts[seq_name],
                                    eval_params=self.eval_params)

        seq_df = postprocess.postprocess_trajectories()


        if output_path is not None:
            self._save_results_to_file(seq_df, output_path)

        return seq_df, constr_sr

    def _save_results_to_file(self, seq_df, output_file_path):
        """
        Stores the tracking result to a txt file, in MOTChallenge format.
        """


        final_out = seq_df[TRACKING_OUT_COLS].sort_values(by=['frame', 'ped_id'])
        final_out.to_csv(output_file_path+'Res', header=True, index=False)
        try:
            final_eval = seq_df[TRACKING_EVAL_COLS].sort_values(by=['frame', 'ped_id'])
            final_eval.to_csv(output_file_path, header=False, index=False)
        except:
            pass



    